/**
 * Copyright (C) 2012 Matteo Tamigi
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, 
 * USA.
*/

#include "PDFSec.hpp"
#include "rc4.h"
#include "md5.h"

static void
printHexString(vector<uint8_t> str) {
  for(int i=0;i<str.size();i++)
    printf("%x ",str[i]);
  printf("\n");
}

//Costanti di cifratura per inizializzare i vari algoritmi di cifratura
const uint8_t pad[] = {
  0x28,0xBF,0x4E,0x5E,0x4E,0x75,0x8A,0x41,
  0x64,0x00,0x4E,0x56,0xFF,0xFA,0x01,0x08,
  0x2E,0x2E,0x00,0xB6,0xD0,0x68,0x3E,0x80,
  0x2F,0x0C,0xA9,0xFE,0x64,0x53,0x69,0x7A
};


//Funzioni per la cifratura/decifratura delle varie versioni

bool PDFSec :: isUserPasswordRev2 () {
    vector<uint8_t> dec;
    dec = md5(key);                                     //Cifro la key con l'md5
    dec.erase(dec.begin()+5,dec.end());                 //Prendo le prime 5 cifre
    dec = Rc4( infoPdf->u_string , dec );               //Cifro la u_str con le 5 cifre prese prima
    //Devo ottenere in dec il pad
    for(unsigned i = 0; i < 32; i++)
        if(dec[i] != pad[i])
            return false;
    return true;
}

bool PDFSec :: isUserPasswordRev3 () {
    vector<uint8_t> dec;
    vector<uint8_t> tmpkey;
    vector<uint8_t> test;
    vector<uint8_t> cifr;
    int length = infoPdf->length/8;
    //Genero la chiave di test (diversa dal pad)
    //md5 ( pad + fileID )
    for(unsigned i = 0; i < 32; i++)
        cifr.push_back(pad[i]);
    for(unsigned i = 0; i < infoPdf->fileID.size(); i++)
        cifr.push_back(infoPdf->fileID[i]);
    cifr = md5(cifr);
    
    //Cifro la key per ottenere cifr
    dec = md5(key);                         //Cifro la key con l'md5
    dec = md5_50(dec);                      //Cifro in md5 altre 50 volte (nel dubbio :D)
    //Genero il vettore per la chiave temporanea
    for(int i = 0; i < length; i++)
        tmpkey.push_back(0x00);
    //La chiave da usare  a 128 bit, mi servono solo i primi 16 byte della u_string
    for(int i = 0; i < 16; i++)
        test.push_back(infoPdf->u_string[i]);
    //Cifro con l'RC4
    for(int i = 19; i >= 0; --i) {
        for(unsigned j = 0; j < length; ++j)
            tmpkey[j] = dec[j] ^ i;
        test = Rc4(test,tmpkey);
    }
    //Confronto il risultato atteso cifr con la user_pswd cifrata
    for(unsigned i = 0; i < 16; i++)
        if(test[i] != cifr[i])
            return false;
    return true;
}


//Funzioni della classe

void PDFSec :: initCracking ( const string usr_pswd = "" ) {
    //Stabilisco la versione della cifratura ed effettuo le operazioni preliminari:
    //-Verifico che i parametri in ingresso siano corretti
    //-Se esiste la user password, verifico che sia quella corretta
    //-Genero il dizionario per le password di bruteforcing
    //------------------------------------------------------------------------//
    
    /*--------------------------------------------------------------------------
    Standard di cifratura PDF:
    V: 1    R: 1 - 3    RC4 a 40-bit con parole da 32 caratteri Latin-1
            R: 4        RC4 a 128-bit con parole da 32 caratteri Latin-1
            R: 5        come la R4 solo diverso metodo di cifratura (NO SPECIFICHE)
            R: 6 - 7    AES-128 con parole da 32 caratteri Latin-1 (NO SPECIFICHE)
            R: 7 ext 3  AES-256 con parole da 127 UTF-8 bytes
                 ext 8  come ext 3 ma diverso metodo di cifratura
    --------------------------------------------------------------------------*/
    
    //Riempio la chiave da cifrare
    // password_padded + o_string + permissions + fileID [ (se la revision  > 3) + encryptMetaData]
    
    //Se c' la user password la inserisco e la utilizzo
    if(usr_pswd != "") {
        userPswd = usr_pswd;
        isUserPswd = true;
        if( 32 < userPswd.length())
            userPswd.erase(userPswd.begin()+32,userPswd.end());    //Se la password  superiore ai 32 caratteri, la taglio
        //Riempio la chiave di cifratura con la user_pswd
        for(unsigned i = 0; i < userPswd.length(); i++)
            key.push_back((uint8_t)userPswd[i]);
    } else 
        isUserPswd = false;
    
    for(unsigned i = 0; i < 32-userPswd.length(); i++)
        key.push_back(pad[i]);
    for(unsigned i = 0; i < 32; i++)
        key.push_back(infoPdf->o_string[i]);
    key.push_back(infoPdf->permissions & 0xff);
    key.push_back((infoPdf->permissions >> 8) & 0xff);
    key.push_back((infoPdf->permissions >> 16) & 0xff);
    key.push_back((infoPdf->permissions >> 24) & 0xff);
    for(unsigned i = 0; i < infoPdf->fileID.size(); i++)
        key.push_back(infoPdf->fileID[i]);
    if(infoPdf->revision > 3 && !infoPdf->encryptMetaData) {
        key.push_back(0xff);
        key.push_back(0xff);
        key.push_back(0xff);
        key.push_back(0xff);
    }
    //Inizializzo il sistema di generazione password
    pswd.SetCharset(STDCHARSET);
    pswd.SetState(1,6);
}

string PDFSec :: crackUserPassword () {
    vector<uint8_t> pswd_padd;
    char * parola;
    //Verifico che sia stato inizializzato l'ambiente di cracking
    if( key.empty() ) {
        cerr << "WARNING: Cracking workspace isn't initialized! Initializing..." << endl;
        initCracking();
    }
    
    
    //Visto che la versione (ottobre 2012)  per tutte le cifrature uguali (pari a 1)
    //verifico solamente la revisione ( parametro R presente in infoPdf->revision )
    //Verifico se la user_password sia quella nulla (per sicurezza)
    switch ( infoPdf->revision ) {
        case 2:
    
            if( !isUserPswd && isUserPasswordRev2() )
                return "";
            //Faccio partire il ciclo di cracking
            while ( 1 ) {
                //Recupero la password da provare
                parola = pswd.GetPswd();
                pswd_padd = generatePaddedPassword(parola);
                for(unsigned i = 0; i < 32; i++)
                    key[i] = pswd_padd[i];
                if ( isUserPasswordRev2() )
                    return string(parola);
                pswd.SetNextPswd();
            }
            break;
        case 3:
            if( isUserPasswordRev3() )
                return "";
            while ( 1 ) {
                //Recupero la password da provare
                parola = pswd.GetPswd();
                pswd_padd = generatePaddedPassword(parola);
                for(unsigned i = 0; i < 32; i++)
                    key[i] = pswd_padd[i];
                if ( isUserPasswordRev3() )
                    return string(parola);
                pswd.SetNextPswd();
            }
            break; 
    }
}

//La funzione recupera la stringa e un vettore e genera la padded password
vector<uint8_t> PDFSec :: generatePaddedPassword ( char *str ) {
    vector<uint8_t> v;
    int len = strlen(str);
    for(unsigned i = 0; i < len; i++)
        v.push_back((uint8_t)str[i]);
    for(unsigned i = 0; i < 32 - len; i++)
        v.push_back(pad[i]);
    return v;
}
